# I know this take too much time imma try optimizing it later on

from json import load
from .. import (
  utils,
  classes
)
from fernet import Fernet
from types import CodeType
import base64
import ast

# Optimized version of the real function :3
def DeobfuscateWeb(ciphertext, key):
  decrypted = {char: i for i, char in enumerate(key)}
  return bytes(decrypted[char] for char in ciphertext)

def Decrypt(data, key) -> str:
  firstLayer = Fernet(key).decrypt(base64.b85decode(data))
  decodedFL = base64.b85decode(firstLayer)
  secondLayer = bytes((i % 94 + 32) for i in range(len(decodedFL)))
  result = bytes(a ^ b for a, b in zip(decodedFL, secondLayer))
  return result.decode()


def findMainPyc(struct: dict[str, bytes]):
  for name, content in struct.items():
    if b'check_debug' in content:
      return content

def main(file: classes.Stub) -> dict:
  
  mainPyc = findMainPyc(file.struct)
  
  loaded = utils.loadPyc(mainPyc, file.version)[0]
  
  # First func is found after the '__main__' const
  func1 = loaded.co_consts[loaded.co_consts.index('__main__') - 2]
  
  # I know hard coding index is not good but :<
  func2 = func1.co_consts[-3]

  strings = (i for i in func2.co_consts if isinstance(i, str))
  ciphertext = next(const for const in strings if len(const) > 500)

  if not ciphertext:
    raise Exception('Couldn\'t find the ciphertext for TrapStealer')
  
  # Still hard coding (idek if it work for every grabbers)
  func3 = func1.co_consts[2]
  
  # I don't have any excuse..
  key = func3.co_consts[-2]
 
  if not key:
    raise Exception('Couldn\'t find the key for TrapStealer')
  
  code = Decrypt(ciphertext, key)
  
  
  encwb, keywb = (i.value for i in utils.getVar(code, 'webhook').elts)
  
  config = {
    "logfile": utils.getVarConst(code, 'logfile'),
    "debug": utils.getVarConst(code, 'debug'),
    "FakeWebhook": utils.getVarConst(code, 'FakeWebhook'),
    "Fakegen": utils.getVarConst(code, 'Fakegen'),
    "FakeCCgen": utils.getVarConst(code, 'FakeCCgen'),
    "FakeError": utils.getVarConst(code, 'FakeError'),
    "schedule": utils.getVarConst(code, 'schedule'),
    "injection": utils.getVarConst(code, 'injection'),
    "Startup": utils.getVarConst(code, 'Startup'),
    "antidebugging": utils.getVarConst(code, 'antidebugging'),
    "DiscordStop": utils.getVarConst(code, 'DiscordStop'),
    "OneTimeSteal": utils.getVarConst(code, 'OneTimeSteal'),
    "melter": utils.getVarConst(code, 'melter'),
    "crasher": utils.getVarConst(code, 'crasher'),
    "hidewindow": utils.getVarConst(code, 'hidewindow'),
    "changebio": utils.getVarConst(code, 'changebio'),
    "Drive": utils.getVarConst(code, 'Drive'),
    "close_proc": utils.getVarConst(code, 'close_proc'),
    "ArchiStealer": utils.getVarConst(code, 'ArchiStealer'),
    
    # WEBSITE UPLOAD
    "Gofile": utils.getVarConst(code, 'Gofile'),
    "fileio": utils.getVarConst(code, 'fileio'),
    "catbox": utils.getVarConst(code, 'catbox'),
    
    # TRAP EXTENSION
    "trap_extension": utils.getVarConst(code, 'trap_extension'),
    "Iban_Stealer": utils.getVarConst(code, 'Iban_Stealer')
  }
  
  
  if not encwb:
    raise Exception('Failed to found the encrypted webhook for TrapStealer')
  
  return {'webhooks': [DeobfuscateWeb(encwb, keywb).decode()], 'config': config}


from .utils import mergeAdd
from .classes import Stub
from .methods import *
from . import methods
import warnings
import io

warnings.filterwarnings("ignore", category=SyntaxWarning)

def getMethod(methodName):
  """
  Get a method from the methods folder by name (Return Any method if not found)

  Args:
      methodName (str): The method name

  Returns:
      Function: The main function of the method
  """
  module = getattr(methods, methodName, None)
  if not module:
    if not __debug__:
      print('DEBUG: Couldn\'t identify the stealer. proceeding using general method')
    return getMethod('Any')
  func = getattr(module, 'main', None)

  return func

def loads(buffer: bytes) -> Stub:
  """
  Get a stub object for a given file from buffer (raw pyc/exe) 

  Args:
      buffer (bytes): The buffer for the file

  Returns:
      Stub: The stub object of the file
  """
  return Stub(FileName = 'placeholder', FileContent = buffer, FileSize = len(buffer), fp = io.BufferedReader(buffer), isExe = buffer.startswith(b'MZ'))

def load(fp: io.BufferedReader) -> Stub:
  """
  Get a stub object for a given file from name

  Args:
      FileName (str): The Name of the file to load

  Returns:
      Stub: The stub object of the file
  """
  content = fp.read()
  fp.seek(0)
  return Stub(FileName = 'placeholder', FileContent = content, FileSize = len(content), fp = fp, isExe = content.startswith(b'MZ'))

def decompile(Object: str | Stub) -> dict:
  """
  Decompile a Pyc or Py file

  Args:
      Object (str/Stub): FileName or Stub object to decompile

  Returns:
      dict: The config Of The Grabber
  """
  
  
  if isinstance(Object, str):
    with open(Object, 'rb') as fp:
      content = fp.read()
      Object = Stub(FileName = Object, FileContent = content, FileSize = len(content), fp = fp, isExe = content.startswith(b'MZ'))

  result = {}
   
  if not Object.isExe:
    found = getMethod('Any')(Object)
    result = found
    
  return mergeAdd(result, getMethod(Object.getType())(Object))


from .. import (
  utils,
  classes
)
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
import base64

def DecryptCode(key, tag, nonce, _input):
  cipher = Cipher(algorithms.AES(key), modes.GCM(nonce, tag))
  decryptor = cipher.decryptor()
  decryptor = decryptor.update(_input) + decryptor.finalize()
  return decryptor.decode(errors = 'ignore')

def identifyPyc(struct: dict[str, bytes]):
  for filename, content in struct.items():
    if b'DecryptString' in content:
      return content
  return None

def decryptFromPlain(code: str):
  key = utils.getFuncCallArg(code, 'key')
  tag = utils.getFuncCallArg(code, 'tag')
  nonce = utils.getFuncCallArg(code, 'nonce')
  ciphertext = utils.getFuncCallArg(code, 'encrypted_data')
  
  return DecryptCode(
    base64.b64decode(key),
    base64.b64decode(tag),
    base64.b64decode(nonce),
    base64.b64decode(ciphertext)
  )

def main(file: classes.Stub) -> dict:
  
  stub = file.struct.get('Stusb', identifyPyc(file.struct))
  
  if not stub:
    raise Exception('Couldn\'t find the stub pyc for ExelaV2')
  
  loaded = utils.loadPyc(stub, file.version)[0]

  index = None
  
  # Find the encrypted code index
  for idx in range(len(loaded.co_consts)):
    const = loaded.co_consts[idx]
    
    if not isinstance(const, str):
      continue
    
    if len(const) > 1000:
      index = idx
      break
  
  if not index:
    raise Exception('Couldn\'t find the encrypted const for ExelaV2')
  
  (key, tag, nonce, ciphertext) = loaded.co_consts[(index - 3) : (index + 1)]
  
  code = DecryptCode(
    base64.b64decode(key),
    base64.b64decode(tag),
    base64.b64decode(nonce),
    base64.b64decode(ciphertext)
  )
  
  # Idk If There Is Always Multiple Iteration But Handle It Anyways
  while 'DecryptString(' in code:
    code = decryptFromPlain(code)
    
  discord_injection = utils.getFuncCallArg(code, 'discord_injection')
  Anti_VM = utils.getFuncCallArg(code, 'Anti_VM')
  StealFiles = utils.getFuncCallArg(code, 'StealFiles')
  
  
  return {'webhooks': utils.getWebhooks(code), 'config': {'discord_injection': discord_injection, 'Anti_VM': Anti_VM, 'StealFiles': StealFiles}}


from .extract import (
  extract
)
from .utils import get_version_from_magics
from typing import *
import io

class Stub:
  """
  The stub class to handle .pyc file and .exe file (for easier extraction)
  """
  def __init__(self, FileName: str, FileContent: dict, FileSize: int, fp, gtype: str = None, fstruct: dict = None, header: bytes = None, isExe: bool = True) -> None:
    if not fstruct:
      fstruct = {}
    if not gtype:
      gtype = ''  
    if not header:
      header = b''
    
    self.name: str = FileName
    self.content: bytes = FileContent
    self.size: int = FileSize
    self.type: Optional[str] = gtype
    self.struct: Optional[dict[str, bytes]] = fstruct
    self.header: bytes = header
    self.type: str = None
    self.isExe: bool = isExe
    self.version: tuple[int, int]
    self.fp = fp
    
  def generateStruct(self) -> None:
    """
    Decompile the pyinstaller package to make a struct
    """
    if not self.isExe:
      self.struct[self.name] = self.content
      self.version = get_version_from_magics(self.content[:4])
      return
    self.struct, self.version = extract(io.BytesIO(self.content))

    
  def getType(self) -> str:
    """
    Get the type of the grabber

    Returns:
        str: The type
    """
    if self.type:
      return self.type
    
    if not self.struct:
      self.generateStruct()
    # self.type = 'BlankGrabber' if 'blank.aes' in self.struct else \
    #             'LunaGrabberV2' if 'luna.aes' in self.struct or 'bound.luna' in self.struct else \
    #             'CrealGrabber' if 'Creal' in self.struct or 'creal' in self.struct else \
    #             'LunaGrabber' if 'main-o' in self.struct else \
    #             'Pysilon' if 'source_prepared' in self.struct else None
    
    self.type = 'BlankGrabber' if 'blank.aes' in self.struct else \
            'CrealGrabber' if 'Creal' in self.struct or 'creal' in self.struct else \
            'LunaGrabber' if 'main-o' in self.struct else \
            'Pysilon' if 'source_prepared' in self.struct else None

    if self.type:
      return self.type
    
    for filename, content in self.struct.items():

      # self.type = 'TrapStealer' if b'python -m pip install crypto' in content and b'requests' in content else \
      #             'ExelaV2' if b'cryptography.hazmat.primitives.ciphers' in content and b'DecryptString' in content else \
      #             'BCStealer' if b'blackcap' in content else \
      #             'CStealer' if b'cs.png' in content else \
      #             'Empyrean' if b'__CONFIG__' in content else \
      #             'OakGrabber' if b'oakgrabber' in content else \
      #             'BLXStealer' if b'blxstealer' in content else \
      #             'LunaGrabber' if b'tkcolorpickerr' in content else \
      #             'RoseStealer' if b'Jrose-stealer' in content else \
      #             'HawkishEyes' if b'Hawkish-Eyes' in content else \
      #             'NiceRAT' if b't.me/NiceRAT' in content else \
      #             'PlainBlankGrabber' if b'Blank Grabber' in content else None
      self.type = 'TrapStealer' if b'detect_debugger_timing' in content else \
            'RedTigerStealer' if b'RedTiger Ste4ler' in content else \
            'BCStealer' if b'blackcap' in content else \
            'CStealer' if b'cs.png' in content else \
            'Pysilon' if b'PySilon' in content else \
            'ExelaV2' if b'cryptography.hazmat.primitives.ciphers' in content and b'DecryptString' in content else \
            'Empyrean' if b'__CONFIG__' in content else \
            'LunaGrabber' if b'tkcolorpickerr' in content else \
            'HawkishEyes' if b'Hawkish-Eyes' in content else \
            'NiceRAT' if b't.me/NiceRAT' in content else \
            'PlainBlankGrabber' if b'Blank Grabber' in content else None
                  
      if self.type:
        return self.type

    self.type = 'Unknown'
    
    return self.type


from .. import (
  utils,
  classes
)
from typing import *
from types import *
import zipfile
import base64
import zlib
import xdis
import re
import io

KeyIVRegex = re.compile(rb'stub-oz,([\w\d=]+)|([\d\w+/=]+)c\x03')

def getKeyIVFromLoader(loader: bytes, version: tuple[int, int]) -> tuple[bytes, int]:
  """
  Return Key And IV From Loader-O File

  Args:
      loader (bytes): The BlankGrabber Loader File

  Returns:
      tuple (bytes, int): The Key And IV
  """
  
  loaded = utils.loadPyc(loader, version)[0]
  
  # key, iv are at index 6 and 7 :3
  
  return (base64.b64decode(loaded.co_consts[6]), base64.b64decode(loaded.co_consts[7]))

def findLoader(struct: dict[str, bytes]) -> bytes:
  
  for i, v in struct.items():
    if b'stub-oz' in v:
      return v
  
  return None
  
def main(file: classes.Stub) -> dict:
  """
  Decompile A Compiled BlankGrabber

  Args:
      file (classes.File): `The file to decompile (exe)`

  Returns:
      str: `List Of Webhook/s found`
  """
  if not file.struct:
    file.generateStruct()

  # Get The Encrypted Payload
  loader_o = findLoader(file.struct)
  cipherText = file.struct.get('blank.aes', None)

  if not loader_o or not cipherText:
    raise FileNotFoundError('Tried To Decompile Invalid BlankGrabber')
  
  key, iv = getKeyIVFromLoader(loader_o, file.version)

  try:
    # Decompress the ciphertext (its reversed)
    uncompressed = zlib.decompress(cipherText[::-1])
  except:
    raise Exception('BlankGrabber Invalid cipherText')
  
  zipArchive = utils.AESDecrypt(key, iv, uncompressed)
  
  # Blank Grabber store the stub in a zip file
  with io.BytesIO(zipArchive) as fileBuffer:
    with zipfile.ZipFile(fileBuffer, 'r') as zipBuffer:
      with zipBuffer.open('stub-o.pyc', 'r') as stub:
        stubData = stub.read()
  
  # Deobfuscate the stub
  obfuscatedStub = utils.findLZMA(stubData)
  compiledStub = utils.BlankObfV1(obfuscatedStub)
  
  codeObj, VTuple, IsPypy, Opcode = utils.loadPyc(compiledStub, file.version)
  
  # Get Insts And Remove Cache Cause I Want To :<
  stubBadInsts = xdis.Bytecode(codeObj, Opcode).get_instructions(codeObj)
  stubInsts = filter(lambda x:x.opname.lower() != 'cache', stubBadInsts)
  
  # Find the settings code object
  for inst in stubInsts:
    if inst.opname != 'LOAD_CONST':
      continue
    if not isinstance(inst.argval, xdis.Code13) and not isinstance(inst.argval, CodeType):
      continue
    if inst.argval.co_name != 'Settings':
      continue
    
    setting_obj = inst.argval
    break
  
  if not setting_obj:
    raise Exception('Couldn\'t find the Settings object for BlankGrabber ')
  
  strings = [inst.argval for inst in xdis.Bytecode(setting_obj, Opcode).get_instructions(setting_obj) if inst.opname == 'LOAD_CONST']

  C2_idx = 0
  for i, v in enumerate(strings):
    if utils.getWebhooks(v):
      C2_idx = i
  
  if C2_idx == 0:
    raise Exception('Couldn\'t find the C2 index for BlankGrabber (Could be using telegram that is not yet supported)')

  # Extract full config (ik that's alot)
  (
    C2,
    Mutex,
    PingMe,
    VmProtect,
    Startup,
    Melt,
    UacBypass,
    ArchivePassword,
    HideConsole,
    Debug,
    RunBoundOnStartup,
    CaptureWebcam,
    CapturePasswords,
    CaptureCookies,
    CaptureAutofills,
    CaptureHistory,
    CaptureDiscordTokens,
    CaptureGames,
    CaptureWifiPasswords,
    CaptureSystemInfo,
    CaptureScreenshot,
    CaptureTelegram,
    CaptureCommonFiles,
    CaptureWallets,
    FakeError,
    FakeErrorConfig,
    BlockAvSites,
    DiscordInjection
  ) = strings[C2_idx : C2_idx+28]
  
  return {
    'webhooks': [base64.b64decode(C2).decode()],
    'config': {
      'Mutex': base64.b64decode(Mutex).decode(),
      'PingMe': bool(PingMe),
      'VmProtect': bool(VmProtect),
      'Startup': bool(Startup),
      'Melt': bool(Melt),
      'UacBypass': bool(UacBypass),
      'ArchivePassword': base64.b64decode(ArchivePassword).decode(),
      'HideConsole': bool(HideConsole),
      'Debug': bool(Debug),
      'RunBoundOnStartup': bool(RunBoundOnStartup),
      'CaptureWebcam': bool(CaptureWebcam),
      'CapturePasswords': bool(CapturePasswords),
      'CaptureCookies': bool(CaptureCookies),
      'CaptureAutofills': bool(CaptureAutofills),
      'CaptureHistory': bool(CaptureHistory),
      'CaptureDiscordTokens': bool(CaptureDiscordTokens),
      'CaptureGames': bool(CaptureGames),
      'CaptureWifiPasswords': bool(CaptureWifiPasswords),
      'CaptureSystemInfo': bool(CaptureSystemInfo),
      'CaptureScreenshot': bool(CaptureScreenshot),
      'CaptureTelegram': bool(CaptureTelegram),
      'CaptureCommonFiles': bool(CaptureCommonFiles),
      'CaptureWallets': bool(CaptureWallets),
      'FakeError': bool(FakeError),
      'FakeErrorConfig': FakeErrorConfig,
      'BlockAvSites': bool(BlockAvSites),
      'DiscordInjection': bool(DiscordInjection)
    }
}


from .. import (
  utils,
  classes,
  extract
)
import xdis
import io


def Deobf(comp: xdis.Code3, xor_table: list) -> str:

  # Extract The Values (remove the last one cause not needed)
  values = [round(abs(char)) for char in comp.co_consts][:-1]

  # Group Values Per 2 (Index, Value)
  grouped = zip(values[::2], values[1::2])
  
  # Sort Groups By Index
  sortedGroups = sorted(grouped, key = lambda x: x[0])

  # Decrypt Char By Char With xor_table[index]
  decrypted = ''.join(chr(value ^ xor_table[idx % len(xor_table)]) for idx, value in sortedGroups)
  
  return decrypted


def Extract(loadedPyc: xdis.Code3, version: tuple[int, int]) -> dict:
  
  xor_table = []
  
  pycObject = xdis.Bytecode(loadedPyc, xdis.get_opcode(version, False))

  insts = pycObject.get_instructions(loadedPyc)
  
  # Extract The Xor Table
  for inst in insts:
    if inst.opname == 'LOAD_CONST':
      xor_table.append(round(abs(inst.argval)))
    elif inst.opname ==  'STORE_NAME':
        break
      
  # Get Consts After The Xor Table  
  consts = [const for const in loadedPyc.co_consts[len(xor_table) + 1:] if isinstance(const, xdis.Code3)]

  # Create Config (store webhook first)
  config = [Deobf(consts[1], xor_table)]
  
  config.extend(inst.argval for inst in insts if inst.opname == 'LOAD_CONST' and isinstance(inst.argval, bool))
  
  __CONFIG__ = {
    'antidebug': config[1],
    'browsers': config[2],
    'discordtoken': config[3],
    'startup': config[4],
    'systeminfo': config[5],
  }
  
  return {'webhook': [config[0]], 'config': __CONFIG__}

def main(file: classes.Stub) -> list[str] | list[None]:
  
  pyz = file.struct.get('PYZ-00.pyz', file.struct.get('PYZ-00.pyzMEI', None))
  if not pyz:
    raise Exception('Couldn\'t Find The PYZ-00 For Empyrean Method')
  
  with io.BytesIO(pyz) as fp:
    config = extract.extractPyzFromName(fp, 'config')
    
  if not config:
    raise Exception('Couln\'t Find The Config File For Empyrean Method')
    
  pycTuple = utils.loadPyc(config, file.version)
  loaded = pycTuple[0]
  
  return Extract(loaded, file.version)


def ungrab():
  import sys
  import Ungrabber
  
  if len(sys.argv) > 1:
    for i, v in Ungrabber.decompile(sys.argv[1]).items():
      print(f'{i}: {v}')
      
if __name__ == '__main__':
  ungrab()


from .. import (
  utils,
  classes
)


from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives import padding
from cryptography.hazmat.backends import default_backend
import base64

def findMainPyc(struct: dict[str, bytes]):
  for name, content in struct.items():
    if b'redtiger.shop' in content:
      return content

def DeriveKey(password, salt):
  kdf = PBKDF2HMAC(hashes.SHA256(), 32, salt, 100000, default_backend())
 
  if isinstance(password, str):
    password = password.encode()
    
  return kdf.derive(password)

def Decrypt(ciphertext, key) -> str:
  decoded = base64.b64decode(ciphertext)
  salt = decoded[:16]
  iv = decoded[16:32]
  decoded = decoded[32:]
  derived_key = DeriveKey(key, salt)
  cipher = Cipher(algorithms.AES(derived_key), modes.CBC(iv), backend = default_backend())
  decryptor = cipher.decryptor()
  decrypted_data = decryptor.update(decoded) + decryptor.finalize()
  unpadder = padding.PKCS7(128).unpadder()
  return (unpadder.update(decrypted_data) + unpadder.finalize()).decode()

def main(file: classes.Stub) -> dict:
  
  loaded = utils.loadPyc(findMainPyc(file.struct), file.version)[0]
  
  # index of const below - 1 = key, - 2 = ciphertext
  
  # Use this const cause its unique
  constIdx = loaded.co_consts.index('RedTiger Ste4ler - github.com/loxy0dev/RedTiger-Tools')

  key = loaded.co_consts[constIdx - 1]
  ciphertext = loaded.co_consts[constIdx - 2]
  

  return {'webhooks': [Decrypt(ciphertext, key)], 'config': {}}


from setuptools import setup, find_packages

setup(
    name='ungrabber',
    version='0.0.2',
    license='MIT',
    description='Ungrabber is a python module to automatically decompile and get the C2/Type of almost every known python grabbers',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    entry_points={
        "console_scripts": [
            "ungrab=Ungrabber.cli:ungrab",
        ],
    },
    install_requires=['xdis', 'cryptography', 'pycryptodome'],
    url='https://github.com/lululepu/Ungrabber',
    author='Lululepu',
    author_email='a.no.qsdf@gmail.com'
)  


"""
This is the method template to use you can do anything you want as long as you
return the webhooks as webhooks in the dict (same for configs and token).
When you make a mehtods don't forget to add it to the __init__.py in the methods folder and
always name the file with the method name the same as the grabber name in the detection file (classes.py)

"""

from .. import (
  utils,
  classes
)

def main(file: classes.Stub) -> dict:
  return {'webhooks': [], 'config': {}}


from .. import (
  utils,
  classes
)
import base64

def main(file: classes.Stub) -> dict:

  if file.isExe:
    source_prepared = file.struct.get('source_prepared', None)
  else:
    source_prepared = file.struct[file.name]
  
  if not source_prepared:
    raise Exception('Couldn\'t Find The Source Prepared For Pysilon')
  
  loaded = utils.loadPyc(source_prepared, file.version)[0]

  # In Const List Tokens Are Stored After The 'auto' Const
  tokenIdx = loaded.co_consts.index('auto') + 1

  # Handle if there is multiple webhooks
  if isinstance(loaded.co_consts[tokenIdx], tuple):
    tokens = [base64.b64decode(token[::-1]).decode() for token in loaded.co_consts[tokenIdx]]
  else:
    tokens = [base64.b64decode(loaded.co_consts[tokenIdx][::-1]).decode()]
    
  return {'webhooks': [], 'tokens': tokens}


# TODO: Make the extractor in c++ for better performances

import io
import os
import zlib
import struct
import marshal
from dataclasses import dataclass

pyinstHeaderSign = b'MEI\x0C\x0B\n\x0B\x0E'

@dataclass
class PyinstHeader:
  Signature: bytes
  PackageSize: int
  TOCoffset: int
  TOCsize: int
  PythonVersion: int
  
  @staticmethod
  def parse(fp: io.BufferedReader, headerOffset: int) -> "PyinstHeader":
    
    fp.seek(headerOffset)
    
    header = PyinstHeader(
      fp.read(len(pyinstHeaderSign)),
      struct.unpack('!I', fp.read(4))[0],
      struct.unpack('!I', fp.read(4))[0],
      struct.unpack('!I', fp.read(4))[0],
      struct.unpack('!I', fp.read(4))[0]
    )

    return header

@dataclass
class PyinstEntry:
  Size: int
  Offset: int
  CompressedSize: int
  UncompressedSize: int
  CompressionFlag: int
  type: int
  name: str
  
  @staticmethod
  def parse(fp: io.BufferedReader, overlayOffset: int) -> "PyinstEntry":
    """
    Parse a pyinstaller entry

    Args:
        fp (io.BufferedReader): The main File Pointer
        overlayOffset (int): The offset of pyinstaller overlay

    Returns:
        PyinstEntry: The parsed entry (relative to overlay start)
    """
    size = struct.unpack('!I', fp.read(4))[0]
    
    # TotalSize - ((Size) Size + (Offset) Size + (CompressedSize) Size + (UncompressedSize) Size + (CompressionFlag) Size + (type) Size)
    nameSize = size - (4 + 4 + 4 + 4 + 1 + 1)
  
    entry = PyinstEntry(
      size,
      overlayOffset + struct.unpack('!I', fp.read(4))[0],
      struct.unpack('!I', fp.read(4))[0],
      struct.unpack('!I', fp.read(4))[0],
      struct.unpack('c', fp.read(1))[0],
      struct.unpack('c', fp.read(1))[0],
      ''
    )
    
    try:
      entry.name = fp.read(nameSize).replace(b'\x00', b'').decode()
    except:
      entry.name = 'InvalidName'
    
    return entry

def getHeaderOffset(fp: io.BufferedReader) -> int:
  """
  Find the header of the pyinstaller archive and return the pos

  Args:
      fp (io.BufferedReader): The main File Pointer

  Returns:
      int: The position of the header relative to the file start
  """
  content = fp.read()
  pyInstHeader = content.rfind(pyinstHeaderSign)
  
  return pyInstHeader

def processEntry(fp: io.BufferedReader, entry: PyinstEntry) -> bytes:
  """
  Process a pyinstaller entry and return the decompressed data

  Args:
      fp (io.BufferedReader): The main File Pointer
      entry (PyinstEntry): The entry to process

  Returns:
      bytes: The decompressed data
  """
  if entry.CompressedSize == 0:
    return b''
  
  content = fp.read(entry.CompressedSize)

  if content.startswith(b'PYZ'):
    return content

  return zlib.decompress(content)

def extract(fp: io.BufferedReader) -> tuple[dict, tuple[int, int]]:
    """
    Main Extract function

    Args:
        fp (io.BufferedReader): The main File Pointer

    Raises:
        Exception: If a bad pyinstaller archive is detected

    Returns:
        tuple[dict, tuple[int, int]]: A tuple of the archive structure and the version tuple 
    """
    fp.seek(0, os.SEEK_SET)
    headerOffset = getHeaderOffset(fp)
    
    if headerOffset == -1:
      raise Exception('Invalid pyinstaller archive')
    
    filesize = fp.seek(0, os.SEEK_END)
    
    fp.seek(headerOffset + 24)
    
    # Extract The Pyinstaller Header
    pyInstHeader = PyinstHeader.parse(fp, headerOffset)
    
    currPos = fp.tell()
    addBytes = fp.seek(0, os.SEEK_END) - currPos - 64


    version = (pyInstHeader.PythonVersion//100, pyInstHeader.PythonVersion % 100)

    # Get The Overlay Offset Relative To The File Start
    overlayOffset = filesize - pyInstHeader.PackageSize - addBytes

    fp.seek(overlayOffset + pyInstHeader.TOCoffset)
    
    bytesRead = 0
    TOC: list[PyinstEntry] = []
    
    # Extract TOC Objects
    while bytesRead < pyInstHeader.TOCsize:
      entry = PyinstEntry.parse(fp, overlayOffset)
      bytesRead += entry.Size
      TOC.append(entry)

    fp.seek(overlayOffset)
    
    result = {entry.name: processEntry(fp, entry) for entry in TOC}
    
    return (result, version)

def getPyzTOC(fp: io.BufferedReader):
  
  # Skip The Header
  fp.seek(8)
  
  (tocPosition, ) = struct.unpack('!i', fp.read(4))
  
  fp.seek(tocPosition, os.SEEK_SET)
  
  TOC = marshal.load(fp)
  
  return TOC

def extractPyzFromName(fp: io.BufferedReader, filename: str):

  TOC = getPyzTOC(fp)
  
  entry = next((x for x in TOC if x[0] == filename), None)

  if not entry:
    return None
  
  (name, (ispkg, pos, length)) = entry
  
  fp.seek(pos)
  
  return zlib.decompress(fp.read(length))

def extractPyz(fp: io.BufferedReader) -> dict[str, bytes]:
  
  TOC = getPyzTOC(fp)
  
  firstEntryPos = TOC[0][1][1]
  
  fp.seek(firstEntryPos)
  
  result = {}
  
  for entry in TOC:
    (name, (ispkg, pos, length)) = entry
    result[name] = zlib.decompress(fp.read(length))
  
  return result


from .. import (
  utils,
  classes
)

def DeobfBlank(file: bytes):
  code = utils.findLZMA(file)
  return utils.BlankObfV1(code.decode())

def scanFile(file: bytes) -> list:
  """
  This function scan a file for plain/encoded webhook and obfuscation

  Args:
      file (bytes): The file to analyze

  Returns:
      list: The webhooks found
  """
  founds = []
  
  founds.extend(utils.getWebhooks(file))  

  obfuscator = utils.DetectObfuscator(file)
  if not obfuscator:
    return founds

  match obfuscator:
    case 'BlankObf':
      founds.extend(scanFile(DeobfBlank(file)))
      
  return founds

def main(file: classes.Stub) -> dict:
  
  found = []
  
  for name, content in file.struct.items():
    # Skip useless files
    if name.endswith(('.pyd','.dll','MEI')):
      continue
    else:
      found.extend(scanFile(content))

  return {'webhooks': found}


"""
MIT License

Copyright (c) 2024 lululepu

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from contextlib import redirect_stderr
import re
from Crypto.Cipher import AES
import xdis.codetype
from typing import *
from . import regs
from types import *
import functools
import base64
import codecs
import struct
import lzma
import xdis
import ast
import sys
import io

LZMASign = b'\xFD\x37\x7A\x58\x5A\x00'
moduleStartByte = (b'\xE3', b'\x63')

versions = {
  (50823, 50823): (2, 0),
  (60202, 60202): (2, 1),
  (60717, 60717): (2, 2),
  (62011, 62021): (2, 3),
  (62041, 62061): (2, 4),
  (62071, 62131): (2, 5),
  (62151, 62161): (2, 6),
  (62171, 62211): (2, 7),
  (3000, 3131): (3, 0),
  (3141, 3151): (3, 1),
  (3160, 3180): (3, 2), # Supported
  (3190, 3230): (3, 3), # Supported
  (3250, 3310): (3, 4), # Supported
  (3320, 3351): (3, 5), # Supported
  (3360, 3379): (3, 6), # Supported
  (3390, 3399): (3, 7), # Supported
  (3400, 3419): (3, 8), # Supported
  (3420, 3429): (3, 9), # Supported
  (3430, 3449): (3, 10), # Supported
  (3450, 3499): (3, 11), # Supported
  (3500, 3549): (3, 12), # Supported
  (3550, 3599): (3, 13), # Supported
  (3600, 3649): (3, 14), # Supported
  (3650, 3699): (3, 15),
  
}
PY310 = b'\x6F\x0D\x0D\x0A'+b'\00'*12
PY311 = b'\xA7\x0D\x0D'+b'\x00'*13
PY312 = b'\xCB\x0D\x0D\x0A'+b'\00'*12
PY313 = b'\xF3\x0D'+b'\00'*14



CODE_REF = b'\xE3'
CODE = b'\x63' # CODE_REF & ~128

def magic_to_int(magic: bytes) -> int:
  return struct.unpack("<Hcc", magic)[0]

def get_version_from_magics(magicBytes: bytes) -> tuple[int, int] | None:
  magicInt = magic_to_int(magicBytes)
  for i, v in versions.items():
    if magicInt >= i[0] and magicInt <= i[1]:
      return v
  return None

def setHeader(pyc: bytes, header: bytes) -> bytes:
  """
  Set Given Header To The Pyc File Safely

  Args:
      pyc (bytes): The Target Pyc
      header (bytes): The Header To Set
      

  Returns:
      bytes: The Pyc With Modified Header
  """
  if pyc.startswith((CODE, CODE_REF)):
    return header + pyc
  
  return header + pyc[16:]


def isValidHeader(pyc: bytes) -> bool:
  """
  Check if a pyc header is valid and supported

  Args:
      pyc (bytes): The pyc to check

  Returns:
      bool: True if header is valid
  """
  
  return pyc.startswith((PY310[2:], PY311[2:], PY312[2:], PY313[2:]))

def getHeader(pymin: int) -> bytes:
  """
  This function give you the pyc header of the given python version

  Args:
      pymin (int): `The python min version`:
      ```
      11 = 3.11 , 12 = 3.12
      ```
  Returns:
      bytes: `The pyc header`
  """

  match pymin:
    case 10:
      return PY310
    case 11:
      return PY311
    case 12:
      return PY312
    case 13:
      return PY313
    case _:
      raise Exception(f'Unsupported version given to function: getHeader')
  

def getWebhooks(content) -> list:
  """
  Return a list of Plain webhooks/B64 Encoded webhoook/C2 found in a strings

  Args:
      content (str): `Text containing a webhook (plain/b64 webhook)`

  Returns:
      list: `List of webhook found (plain)`
  """
  
  content = str(content)
  
  encoded = sum([
    regs.DiscordB64Webhook.findall(content),
    regs.CanaryB64Webhook.findall(content),
    regs.PTBB64Webhook.findall(content),
    regs.DiscordAppB64Webhook.findall(content),
  ], [])
  founds = [base64.b64decode(webhook) for webhook in encoded]
    
  founds.extend(regs.DiscordWebhook.findall(content))
  return founds 

# I use this function to optimize code.
# @functools.lru_cache(maxsize=5) (disabled due to problem with blank)
def walk_cache(code: str):
  return ast.walk(ast.parse(code))

def getVar(code: str, var: str):
  """
  Internal Use Function
  """
  for node in walk_cache(code):
    if isinstance(node, ast.Assign) and any(isinstance(target, ast.Name) and target.id == var for target in node.targets):
      return node.value


def getVarConst(code: str, var: str) -> bool:
  """
  Get an Constant value in a code from name

  Args:
      code (str): `The code to analyse`
      var (str): `The name of the var to retrieve`

  Returns:
      Any: `The value of the var`
  """

  return getVar(code, var).value

def getFuncCallArg(code: str, varname: str) -> str:
  """
  Get a arg from a function call example:
  ```
  code = '''key = base64.b64encode('Test')'''
  print(getFuncCallArg(code, 'key')))
  ```
  `This will return 'Test'`

  Args:
      code (str): `The code to analyse`
      varname (str): `The name of the var to retrieve`

  Returns:
      str: `The arg of the func call`
  """
  
  return getVar(code, varname).args[0].value

def AESDecrypt(key: bytes, iv: bytes, ciphertext: bytes) -> bytes:
  """
  Decrypt AES encrypter ciphertext

  Args:
      key (bytes): `The secret key to use in the symmetric cipher.`
      iv (bytes): `The initialization vector to use for encryption or decryption.`
      ciphertext (bytes): `The piece of data to decrypt.`

  Returns:
      bytes: Decrypted Plain Text
  """
  cipher = AES.new(key, AES.MODE_GCM, nonce=iv)
  decrypted = cipher.decrypt(ciphertext)
  return decrypted

def DetectObfuscator(code: bytes):
  if isinstance(code, str):
    code = code.encode()
  
  if b'___________' in code:
    if findLZMA(code):
      return 'BlankObf'
  
  return None

def BlankObfV1(code: str) -> str:
  """
  Deobfuscate BlankObfV1 Plain Obfuscation

  Args:
      code (str): `The code to deobfuscate`

  Returns:
      str: `The deobfuscated code`
  """
  ____ = getVarConst(code, '____')
  _____ = getVarConst(code, '_____')
  ______ = getVarConst(code, '______')
  _______ = getVarConst(code, '_______')
  deobfuscated = base64.b64decode(codecs.decode(____, 'rot13')+_____+______[::-1]+_______)
  content = deobfuscated
  return content

def mergeAdd(dict1: dict, dict2: dict) -> dict:
  """
  Merge 2 dict but add the already existing index instead of replacing

  Args:
      dict1 (dict): First dict
      dict2 (dict): Second dict

  Returns:
      dict: The merged dict
  """
  
  result = dict1
  
  for i, v in dict2.items():
    if result.get(i, False):
      result[i] += v
    else:
      result[i] = v
      
  return result

def findLZMA(content: bytes) -> bytes:
  """
  Find An Lzma Signature And Return The Decompressed Content

  Args:
      content (bytes): The Buffer To Search

  Returns:
      bytes: The Decompressed LZMA
  """
  return lzma.decompress(LZMASign + content.split(LZMASign)[-1])

def loadPyc(pyc: bytes, version: tuple[int, int]) -> tuple[xdis.Code3, tuple[int,int], bool, ModuleType]:
  """
  Load Pyc With Header Or Not

  Args:
      pyc (bytes): `The Target Pyc`

  Returns:
      tuple (xdis.codetype.CodeBase, tuple[int,int], bool, ModuleType): `A Tuple Of Info (CodeObj, VersionTuple, isPypy, OpCode)`
  """
  
  if not isValidHeader(pyc):

    pyc = setHeader(pyc, getHeader(version[1]))
    
  loaded = xdis.load_module_from_file_object(io.BytesIO(pyc))
  
  version_tuple = loaded[0]
  code_obj = loaded[3]
  ispypy = loaded[4]
  opcode = xdis.get_opcode(version_tuple, ispypy)
  
  return (code_obj, version_tuple, ispypy, opcode)


import re

# TelegramToken = re.compile(r'\d+:[a-zA-Z\d_-]+\$(?:-)?\d+')

DiscordWebhook = re.compile(r'https:\/\/(?:canary\.)?(?:ptb\.)?discord(?:app)?\.com\/api\/webhooks\/\d+\/[A-Za-z0-9_-]+')
CanaryB64Webhook = re.compile(r'aHR0cHM6Ly9jYW5hcnkuZGlzY29yZC5jb20vYXBpL3dlYmhvb2tz[a-zA-Z\d]+(?:={1,2})?')
DiscordAppB64Webhook = re.compile(r'aHR0cHM6Ly9kaXNjb3JkYXBwLmNvbS9hcGkvd2ViaG9va3M[a-zA-Z\d]+(?:={1,2})?')
DiscordB64Webhook = re.compile(r'aHR0cHM6Ly9kaXNjb3JkLmNvbS9hcGkvd2ViaG9va3Mv[a-zA-Z\d]+(?:={1,2})?')
PTBB64Webhook = re.compile(r'aHR0cHM6Ly9wdGIuZGlzY29yZC5jb20vYXBpL3dlYmhvb2tz[a-zA-Z\d]+(?:={1,2})?')
